﻿using System;

namespace QuantRiskLib
{
    ///Source: www.risk256.com
    ///
    ///References:
    ///Miller, Michael B. 2012. Mathematics and Statistics for Financial Risk Management. New York: John Wiley & Sons.
    ///Chapter 4

    public class Distributions
    {
        /// <summary>
        /// Returns the probability of an event occuring in k out of n trials, given 
        /// that the probability of the event occuring in any one trial is p.
        /// </summary>
        /// <param name="p">Probability that the event occurs in any one trial.</param>
        /// <param name="n">Number of trials.</param>
        /// <param name="k">Number of times the event occurs in n trials.</param>
        /// <returns></returns>
        public static double BinomialProbabilityDensityFunction(double p, int n, int k)
        {
            return MMath.Combinations(n, k) * Math.Pow(p, k) * Math.Pow(1 - p, n - k);
        }

        /// <summary>
        /// Chi-squared probability density function.
        /// </summary>
        /// <param name="x">The value at which the PDF is evaluated.</param>
        /// <param name="k">Degress of freedom, or number independent standard normal distributions.</param>
        /// <returns></returns>
        public static double ChiSquaredProbabilityDensityFunction(double x, int k)
        {
            double g = MMath.GammaFunction(0.5 * k);
            double a = 1.0 / (Math.Pow(2, 0.5 * k) * g);
            return a * Math.Pow(x, 0.5 * k - 1.0) * Math.Exp(-0.5 * x);
        }

        /// <summary>
        /// The Irwin-Hall distribution results from the sum on n independent standard uniform variables
        /// </summary>
        /// <param name="x">The value at which to evaluate the distribution.</param>
        /// <param name="n">The number of standard uniform variables.</param>
        public static double IrwinHallProbabilityDensityFunction(double x, int n)
        {
            if (x < 0 || x > n) return 0.0;
            double d = 0;
            for (int i = 0; i <= n; i++)
                d += Math.Pow(-1, i) * MMath.Combinations(n, i) * Math.Pow(x - i, n - 1) * Math.Sign(x-i);
            d *= 0.5;
            d /= MMath.Factorial(n - 1);
            return d;
        }

        /// <summary>
        /// Returns the PDF of the lognormal distribution.
        /// </summary>
        /// <param name="x">Value at which the distribution is evaluated.</param>
        /// <param name="mu">This is *not* the mean. See Miller (2012).</param>
        /// <param name="sigma">This is *not* the standard deviation. See Miller (2012).</param>
        public static double LogNormalProbabilityDensityFunction(double x, double mu, double sigma)
        {
            const double SqrtTwoPiInv = 0.398942280401433;
            double z = (Math.Log(x) - mu) / sigma;
            return SqrtTwoPiInv * Math.Exp(-0.5 * z * z) / (x * sigma);
        }

        /// <summary>
        /// Returns the PDF of the normal distribution.
        /// </summary>
        /// <param name="x">Value at which the distribution is evaluated.</param>
        /// <param name="mean">Mean of the distribution.</param>
        /// <param name="standardDeviation">Standard deviation of the distribution.</param>
        public static double NormalProbabilityDensityFunction(double x, double mean, double standardDeviation)
        {
            const double SqrtTwoPiInv = 0.398942280401433;
            double z = (x - mean) / standardDeviation;
            return SqrtTwoPiInv * Math.Exp(-0.5 * z * z) / standardDeviation;
        }

        /// <summary>
        /// Returns the CDF of the normal distribution.
        /// </summary>
        /// <param name="x">Value at which the distribution is evaluated.</param>
        /// <param name="mean">Mean of the distribution.</param>
        /// <param name="standardDeviation">Standard deviation of the distribution.</param>
        public static double NormalCumulativeDistributionFunction(double x, double mean, double standardDeviation)
        {
            double z = (x - mean) / standardDeviation;
            return StandardNormalCumulativeDistributionFunction(z);
        }

        /// <summary>
        /// Returns the inverse of the CDF of the normal distribution.
        /// </summary>
        /// <param name="p">Cumulative probability of the distribution. 0 &lt;= p &gt;= 1.</param>
        ///<param name="mean">Mean of the distribution.</param>
        /// <param name="standardDeviation">Standard deviation of the distribution.</param>
        public static double NormalCumulativeDistributionFunctionInverse(double p, double mean, double standardDeviation)
        {
            double x = StandardNormalCumulativeDistributionFunctionInverse(p);
            return standardDeviation * x + mean;
        }

        /// <summary>
        /// The PDF of the Poisson distribution.
        /// The probability of n iid events occuring in an interval, if the expected number of event is lambda.
        /// </summary>
        /// <param name="lambda">Equals both the mean and variance of the distribution.</param>
        /// <param name="n">The number of events.</param>
        public static double PoissonProbabilityDensityFunction(double lambda, int n)
        {
            return Math.Pow(lambda, n) * Math.Exp(-lambda) / MMath.Factorial(n);
        }

        /// <summary>
        /// Returns the PDF of the standard normal distribution.
        /// </summary>
        /// <param name="x">Value at which the distribution is evaluated.</param>
        public static double StandardNormalProbabilityDensityFunction(double x)
        {
            const double SqrtTwoPiInv = 0.398942280401433;
            return SqrtTwoPiInv * Math.Exp(-0.5 * x * x);
        }

        /// <summary>
        /// Returns the CDF of the standard normal distribution.
        /// </summary>
        /// <param name="x">Value at which the distribution is evaluated.</param>
        public static double StandardNormalCumulativeDistributionFunction(double x)
        {
            //Approimation based on Abramowitz & Stegun (1964)

            if (x < 0)
                return 1.0 - StandardNormalCumulativeDistributionFunction(-x);
            const double b0 = 0.2316419; 
            const double b1 = 0.319381530;
            const double b2 = -0.356563782; 
            const double b3 = 1.781477937;
            const double b4 = -1.821255978;
            const double b5 = 1.330274429;
            double pdf = StandardNormalProbabilityDensityFunction(x);
            double a = 1.0 / (1.0 + b0 * x);
            return 1.0 - pdf * (b1 * a + b2 * Math.Pow(a, 2) + b3 * Math.Pow(a, 3) + b4 * Math.Pow(a, 4) + b5 * Math.Pow(a, 5));
        }

        /// <summary>
        /// Returns the inverse of the CDF of the standard normal distribution.
        /// </summary>
        /// <param name="p">Cumulative probability of the distribution. 0 &lt;= p &gt;= 1.</param>
        public static double StandardNormalCumulativeDistributionFunctionInverse(double p)
        {
            //based on pseudo-code from http://home.online.no/~pjacklam/notes/invnorm/

            if (p < 0 || p > 1.0) throw new System.ArgumentException("Probability must be between 0 and 1");

            const double a1 = -3.969683028665376e+01;
            const double a2 =  2.209460984245205e+02;
            const double a3 = -2.759285104469687e+02;
            const double a4 =  1.383577518672690e+02;
            const double a5 = -3.066479806614716e+01;
            const double a6 =  2.506628277459239e+00;

            const double b1 = -5.447609879822406e+01;
            const double b2 =  1.615858368580409e+02;
            const double b3 = -1.556989798598866e+02;
            const double b4 =  6.680131188771972e+01;
            const double b5 = -1.328068155288572e+01;

            const double c1 = -7.784894002430293e-03;
            const double c2 = -3.223964580411365e-01;
            const double c3 = -2.400758277161838e+00;
            const double c4 = -2.549732539343734e+00;
            const double c5 =  4.374664141464968e+00;
            const double c6 =  2.938163982698783e+00;

            const double d1 =  7.784695709041462e-03;
            const double d2 =  3.224671290700398e-01;
            const double d3 =  2.445134137142996e+00;
            const double d4 =  3.754408661907416e+00;

            const double p_low  = 0.02425;
            const double p_high = 1 - p_low;

            double q, x;
            if (0 < p && p < p_low)
            {
                q = Math.Sqrt(-2*Math.Log(p));
                x = (((((c1*q+c2)*q+c3)*q+c4)*q+c5)*q+c6) / ((((d1*q+d2)*q+d3)*q+d4)*q+1);
                return x;
            }
      
            if (p_low <= p && p <= p_high)
            {
                q = p - 0.5;
                double r = q*q;
                x = (((((a1*r+a2)*r+a3)*r+a4)*r+a5)*r+a6)*q / (((((b1*r+b2)*r+b3)*r+b4)*r+b5)*r+1);
                return x;
            }

            //if(p_high < p && p < 1)
            q = Math.Sqrt(-2*Math.Log(1-p));
            x = -(((((c1 * q + c2) * q + c3) * q + c4) * q + c5) * q + c6) / ((((d1 * q + d2) * q + d3) * q + d4) * q + 1);
            return x;
        }

        /// <summary>
        /// Returns the PDF of Student's t distribution.
        /// </summary>
        /// <param name="x">Value at which the distribution is evaluated.</param>
        /// <param name="k">Degrees of freedom.</param>
        public static double StudentsTProbabilityDensityFunction(double x, int k)
        {
            double a = MMath.GammaFunction(0.5 * k + 0.5);
            double b = Math.Pow(1.0 + (x * x) / k, -0.5 * k - 0.5);
            double c = Math.Sqrt(k * Math.PI) * MMath.GammaFunction(0.5 * k);
            
            return a * b / c;
        }

        /// <summary>
        /// Returns the PDF of the uniform distribution.
        /// </summary>
        /// <param name="x">Value at which the distribution is evaluated.</param>
        /// <param name="min">Minimum value of the distribution.</param>
        /// <param name="maxx">Maximum value of the distribution.</param>
        public static double UniformProbabilityDensityFunction(double x, double min, double max)
        {
            if (max <= min) throw new System.ArgumentException("max must be greater than min");
            if (x < min || x > max) return 0.0;
            return 1.0 / (max - min);
        }

        /// <summary>
        /// Returns the CDF of the uniform distribution.
        /// </summary>
        /// <param name="x">Value at which the distribution is evaluated.</param>
        /// <param name="min">Minimum value of the distribution.</param>
        /// <param name="maxx">Maximum value of the distribution.</param>
        public static double UniformCumulativeDensityFunction(double x, double min, double max)
        {
            if (max <= min) throw new System.ArgumentException("max must be greater than min");
            return (x - min) / (max - min);
        }
    }
}

//Disclaimer
//This code is freeware. The methods are not proprietary. Feel free to use, modify and redistribute. That said, if you plan 
//to use or redistribute give credit where credit is due and provide a link back to Risk256.com (or don't remove the link 
//and references already in the code). The code is intended primarily as an educational tool. No warranty is made as to the 
//code's accuracy. Use at your own risk.
