﻿using System;

namespace QuantRiskLib
{
    ///Source: www.risk256.com
    ///
    ///References:
    ///Miller, Michael B. 2012. Mathematics and Statistics for Financial Risk Management. New York: John Wiley & Sons.
    ///Chapter 1
    
    public class MMath
    {
        /// <summary>
        /// Returns the number of possible permutations of k objects from a set of n object. The order of the chosen objects does matter.
        /// </summary>
        /// <param name="n">Number of objects</param>
        /// <param name="k">Number of objects chosen</param>
        /// <returns></returns>
        public static long Permutations(int n, int k)
        {
            if (n < 0 || k < 0) throw new System.ArgumentException("Combin not defined for negative n or k");
            long permut = 1;
            for (int i = n; i > (n - k); i--)
                permut = checked(permut * i);
            return permut;
        }

        /// <summary>
        /// Returns the number of possible combinations of k objects from a set of n object. The order of the chosen objects does not matter.
        /// </summary>
        /// <param name="n">Number of objects</param>
        /// <param name="k">Number of objects chosen</param>
        /// <returns></returns>
        public static long Combinations(int n, int k)
        {
            if (n < 0 || k < 0) throw new System.ArgumentException("Combin not defined for negative n or k");
            long c = 1;
            for (int i = n; i > (n - k); i--)
                c = checked(c * i);
            long kFact = Factorial(k);
            return c / kFact;
        }

        /// <summary>
        /// Returns n! 
        /// 0! = 1,otherwise n! = n * (n-1) * (n-2) * ... * 2 * 1,
        /// </summary>
        public static long Factorial(int n)
        {
            if (n < 0) throw new System.ArgumentException("Factorial not defined for negative n");
            if (n > 20) throw new System.ArgumentException("Answer will exceed max long");
            long fact = 1;
            for (int i = n; i > 0; i--)
                fact *= i;
            return fact;
        }

        /// <summary>
        /// Integral from 0 to infinity of e^(-t) * t^(z-1) dt
        /// </summary>
        public static double GammaFunction(double z)
        {
            //based on approximation from http://en.wikipedia.org/wiki/Gamma_function

            if (z > 0 && z % 1.0 == 0)
                return Factorial((int)Math.Round(z - 1, 0));
            if (z > 0 && z % 0.5 == 0)
            {
                int n = (int)z;
                return Math.Sqrt(Math.PI) * Factorial(2 * n) / (Math.Pow(4, n) * Factorial(n));
            }

            double sqrtTwoPi = 2.506628274631;
            double d = 1.0 + 1.0 / (12 * z) + 1.0 / (288 * z * z) - 139.0 / (51840 * Math.Pow(z, 3)) - 571.0 / (2488320 * Math.Pow(z, 4));
            double g = Math.Pow(z, z - 0.5) * Math.Exp(-z) * sqrtTwoPi * d;
            return g;
        }
    }
}

//Disclaimer
//This code is freeware. The methods are not proprietary. Feel free to use, modify and redistribute. That said, if you plan 
//to use or redistribute give credit where credit is due and provide a link back to Risk256.com (or don't remove the link 
//and references already in the code). The code is intended primarily as an educational tool. No warranty is made as to the 
//code's accuracy. Use at your own risk.