﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace QuantRiskLib
{
    ///Source: www.risk256.com
    ///
    ///References:
    ///Miller, Michael B. 2012. Mathematics and Statistics for Financial Risk Management. New York: John Wiley & Sons.
    ///Chapters 8 and 10

    public class Regression
    {
        public readonly Matrix Y;
        public readonly Matrix X;
        public Matrix betas;

        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand.</param>
        /// <param name="xData">Data for the independent variable, or regressor.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        public Regression(double[] yData, double[] xData, bool includeConstant)
        {
            Y = new Matrix(yData);
            if (includeConstant)
            {
                double[,] xDataPlus = new double[xData.Length, 2];
                for (int r = 0; r < xDataPlus.GetLength(0); r++)
                {
                    xDataPlus[r, 0] = 1.0;
                    xDataPlus[r, 1] = xData[r];
                }
                X = new Matrix(xDataPlus);
            }
            else
            {
                X = new Matrix(xData);
            }
            if (Y.nRows != X.nRows)
                throw new System.ArgumentException("Dependent and independent data do not have the same number of rows.");

            CalculateStatistic();
        }

        /// <summary>
        /// Populates the data and calculates statistics for an OLS regression analysis.
        /// </summary>
        /// <param name="yData">Data for the dependent variable, or regressand.</param>
        /// <param name="xData">Data for the independent variables, or regressors.</param>
        /// <param name="includeConstant">If true a column of ones is added to the start of the xData array.</param>
        public Regression(double[] yData, double[,] xData, bool includeConstant)
        {
            Y = new Matrix(yData);
            if (includeConstant)
            {
                double[,] xDataPlus = new double[xData.GetLength(0), xData.GetLength(1) + 1];
                for (int r = 0; r < xDataPlus.GetLength(0); r++)
                {
                    xDataPlus[r, 0] = 1.0;
                    for (int c = 1; c < xDataPlus.GetLength(1); c++)
                        xDataPlus[r, c] = xData[r, c - 1];
                }
                X = new Matrix(xDataPlus);
            }
            else
            {
                X = new Matrix(xData);
            }
            if (Y.nRows != X.nRows)
                throw new System.ArgumentException("Dependent and independent data do not have the same number of rows.");

            CalculateStatistic();
        }

        private void CalculateStatistic()
        {
            Matrix XprimeXinv = X.Transpose().Multiply(X).Inverse();
            betas = XprimeXinv.Multiply(X.Transpose()).Multiply(Y);
        }
    }
}

//Disclaimer
//This code is freeware. The methods are not proprietary. Feel free to use, modify and redistribute. That said, if you plan 
//to use or redistribute give credit where credit is due and provide a link back to Risk256.com (or don't remove the link 
//and references already in the code). The code is intended primarily as an educational tool. No warranty is made as to the 
//code's accuracy. Use at your own risk.